from asyncio import base_tasks
import numpy
import cupy

from matplotlib import cm
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvas

class ImageWidget(FigureCanvas):
    def __init__(self, endpoint, parent=None):
        super().__init__(Figure(figsize=(5, 3)))
        self._endpoint = endpoint

        self._image = None
        self._imshow = None

        self._bscan_idxs = []

    def notify_segments(self, dirty):
        self._bscan_idxs.extend(dirty)
        self.update()

class EnFaceImageWidget(ImageWidget):
    def paintEvent(self, e):
        if self._image is None or self._image.shape != self._image_shape():
            if self._endpoint.tensor.shape:
                self._image = numpy.zeros(self._image_shape(), dtype=self._endpoint.tensor.dtype)
            self._imshow = None

        if self._bscan_idxs:
            idxs = sorted(list(self._bscan_idxs))
            self._bscan_idxs.clear()

            self._update_image(self._endpoint, idxs)

        if self._image is not None:
            if self._imshow is None:
                self.figure.clear()
                axes = self.figure.subplots()
                self._imshow = axes.imshow(self._image, interpolation='nearest', vmin=0, vmax=60, aspect=self._image.shape[1] / self._image.shape[0], cmap=cm.turbo)
            else:
                self._imshow.set_array(self._image)

            self.draw()

        super().paintEvent(e)

    def _image_shape(self):
        return tuple(self._endpoint.tensor.shape[:2])

    def _update_image(self, endpoint, bscan_idxs):
        pass

class RasterEnFaceWidget(EnFaceImageWidget):
    def __init__(self, *args, **kwargs):
        title = kwargs.pop('title', 'Raster En Face')

        super().__init__(*args, **kwargs)

        self.setWindowTitle(title)

    def _update_image(self, endpoint, bscan_idxs):
            with endpoint.tensor as volume:
                if isinstance(volume, cupy.ndarray):
                    endpoint.stream.synchronize()
                bscan_idxs = [i for i in bscan_idxs if i < volume.shape[0]]
                if isinstance(volume, cupy.ndarray):
                    self._image[bscan_idxs, ...] = cupy.asnumpy(cupy.max(volume[bscan_idxs, ...], axis=2))
                else:
                    self._image[bscan_idxs, ...] = numpy.max(volume[bscan_idxs, ...], axis=2)

class RadialEnFaceWidget(EnFaceImageWidget):
    def __init__(self, *args, **kwargs):
        title = kwargs.pop('title', 'Radial En Face')

        super().__init__(*args, **kwargs)

        self.setWindowTitle(title)

    def _update_image(self, endpoint, bscan_idxs):
        with endpoint.tensor as volume:
            if isinstance(volume, cupy.ndarray):
                endpoint.stream.synchronize()
                self._image = cupy.asnumpy(cupy.max(volume, axis=2))
            else:
                self._image = numpy.max(volume, axis=2)
class CrossSectionImageWidget(ImageWidget):
    def __init__(self, *args, **kwargs):
        self._fixed = kwargs.pop('fixed', None)
        self._aspect = kwargs.pop('aspect', 1)
        title = kwargs.pop('title', 'Cross-Section')

        super().__init__(*args, **kwargs)

        self.setWindowTitle(title)

    def paintEvent(self, e):
        if self._bscan_idxs:
            idxs = self._bscan_idxs[:]
            self._bscan_idxs.clear()

            self._update_image(self._endpoint, idxs)

        if self._image is not None:
            if self._imshow is None:
                axes = self.figure.subplots()
                self._imshow = axes.imshow(self._image, interpolation='nearest', vmin=0, vmax=30, aspect=self._aspect, cmap=cm.turbo)
            else:
                self._imshow.set_array(self._image)

            self.draw()

        super().paintEvent(e)

    def _update_image(self, endpoint, bscan_idxs):
        if self._fixed is None:
            idx = bscan_idxs[-1]
        elif self._fixed in bscan_idxs:
            idx = self._fixed
        else:
            return

        with endpoint.tensor as volume:
            if isinstance(volume, cupy.ndarray):
                endpoint.stream.synchronize()
                self._image = cupy.asnumpy(volume[idx].T)
            else:
                self._image = volume[idx].T
