# ref: http://www.benjack.io/2018/02/02/python-cpp-revisited.html

import os
import re
import sys
import platform
import subprocess
from pathlib import Path

from distutils.version import LooseVersion
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py

def _dump_vars(title, vars):
    if not vars:
        return

    pad = max([len(k) for k in vars])

    print(f'-- {title}:')
    print()
    for (k, v) in vars.items():
        print(f'    {k: <{pad}} = {v}')
    print()

def cmake_bool(s):
    if s.lower() in ['on', 'yes', 'true', 'y']:
        return True
    try:
        return bool(int(s))
    except ValueError:
        return False

root = Path(__file__).resolve().parent

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=None):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = Path(sourcedir or '').resolve()

class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError('CMake must be installed to build the following extensions: ' + ', '.join(e.name for e in self.extensions))

        cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)', out.decode()).group(1))
        if cmake_version < '3.17.0':
            raise RuntimeError("CMake >= 3.17.0 is required")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        # info needed for CMake variables
        config = 'Debug' if self.debug else 'Release'

        python_var_name = 'Python'
        if sys.version_info.major != 2:
            python_var_name += str(sys.version_info.major)

        cmake_vars = {
            'CMAKE_BUILD_TYPE:STRING': config,

            # force CMake to detect the currently executing Python
            f'{python_var_name}_EXECUTABLE:FILEPATH': sys.executable,

            # let setuptools build the wheel
            'ENABLE_BUILD_PYTHON_WHEEL:BOOL': 'OFF',
        }

        # info needed for environment variables
        bits = 64 if sys.maxsize > 2**32 else 32

        if platform.system() == 'Windows':

            env_vars = {
                # build with clang (MSVC is not supported)
                'CC': 'clang-cl',
                'CXX': 'clang-cl',

                # set x64/x86, suppress known warnings from TBB, and generate debugging information
                'CXXFLAGS': f'-m{bits} -fdiagnostics-absolute-paths -Wno-unused-command-line-argument -Wno-deprecated-volatile -Wno-deprecated-declarations /DTBB_SUPPRESS_DEPRECATED_MESSAGES /Zi',
                'LDFLAGS': '/debug',

                # suppress known warnings and generate debugging information
                'CUDAFLAGS': '-Xcudafe --diag_suppress=base_class_has_different_dll_interface --debug',
            }

            script_suffix = '.bat'

        else:

            env_vars = {
                # build with clang
                'CC': 'clang',
                'CXX': 'clang',

                # set x64/x86 and generate debugging information
                'CXXFLAGS': f'-m{bits} -stdlib=libc++ -fdiagnostics-absolute-paths --debug',

                # generate debugging information
                'CUDACXX': '/usr/local/cuda/bin/nvcc',
                'CUDAFLAGS': '--debug',
            }

            # unix-specific shared object definitions
            cmake_vars.update({
                "CMAKE_BUILD_WITH_INSTALL_RPATH": "ON",
                "CMAKE_POSITION_INDEPENDENT_CODE": "ON",
            })

            script_suffix = '.sh'

        def show(k):
            for s in ['CMAKE', 'VORTEX']:
                if k.startswith(s):
                    return True

        # update existing environment variable
        env = os.environ.copy()
        env.update(env_vars)

        # process the feature specification
        feature_spec = env.get('VORTEX_BUILD_FEATURES', 'WITH_ALAZAR=ON;WITH_CUDA=ON;WITH_DAQMX=ON;WITH_HDF5=ON;WITH_REFLEXXES=ON')

        features = {}
        for spec in feature_spec.split(';'):
            feature, delim, state = spec.strip().partition('=')

            features[feature] = cmake_bool(state) or delim == ''

        # require Python bindings
        features['WITH_PYTHON'] = True
        cmake_vars.update({f: 'ON' if v else 'OFF' for (f, v) in features.items()})

        if not any([k in env for k in ['VORTEX_DISABLE_AUTO_VCPKG', 'CMAKE_TOOLCHAIN_FILE', 'CMAKE_INSTALL_PREFX']]):
            feature2pkgs = {
                'WITH_REFLEXXES': ['reflexxes'],
                'WITH_CUDA': ['cuda', 'cub'],
                'WITH_FFTW': ['fftw3[avx2]'],
                'WITH_PYTHON': ['xtensor-python', 'pybind11'],
                'WITH_HDF5': ['hdf5[cpp]'],
            }

            mandatory_pkgs = ['fmt', 'spdlog', 'xtensor[tbb,xsimd]']

            pkgs = sum([feature2pkgs.get(f, []) for f in features], mandatory_pkgs)

            print('-- Using vcpkg for dependencies')
            # use vcpkg for dependencies automatically

            vcpkg_root = root / 'build' / 'vcpkg'
            if not vcpkg_root.exists():
                subprocess.check_call(['git', 'clone', 'https://github.com/microsoft/vcpkg.git', str(vcpkg_root)])
                subprocess.check_call([str(vcpkg_root / 'bootstrap-vcpkg.{script_suffix}'), '-disableMetrics'])

            vcpkg_args = [str(vcpkg_root / 'vcpkg'), '--clean-after-build', f'--overlay-ports={root / ".vcpkg"}', 'install'] + pkgs
            print('-- vcpkg: ', vcpkg_args)
            subprocess.check_call(vcpkg_args)

            # set CMake to use toolchain
            cmake_vars['CMAKE_TOOLCHAIN_FILE'] = str(vcpkg_root/ 'scripts' / 'buildsystems' / 'vcpkg.cmake')

        _dump_vars('CMake Variables', cmake_vars)
        _dump_vars('Environment Variables', {k: v for (k, v) in env.items() if show(k) or k in env_vars})

        # generate build directory
        build_dir = Path(self.build_temp).resolve()
        build_dir.mkdir(parents=True, exist_ok=True)

        # configure with CMake
        cmake_configure_cmd = [
            'cmake',
            '-G', 'Ninja',
            '-S', str(ext.sourcedir),
            '-B', str(build_dir),
        ] + [f'-D{k}={v}' for (k, v) in cmake_vars.items()]
        print('-- CMake Configure:', cmake_configure_cmd)
        subprocess.check_call(cmake_configure_cmd, env=env)

        # build with CMake
        cmake_build_cmd = ['cmake', '--build', str(build_dir)]
        print('-- CMake Build:', cmake_build_cmd)
        subprocess.check_call(cmake_build_cmd)

        self.prepare_wheel(ext, build_dir)

    def prepare_wheel(self, ext, build_dir, deploy_dependencies=True, generate_stubs=True, deploy_path_resctriction=None):
        # format files for wheel
        extpath = Path(self.get_ext_fullpath(ext.name)).resolve()
        wheel_dir = extpath.parent

        package_name = ext.name
        (wheel_dir / package_name).mkdir(parents=True, exist_ok=True)

        extension_path = build_dir / 'lib' / extpath.name
        vortex_dir = build_dir / 'bin'
        package_root = wheel_dir / ext.name

        # create link to allow import without placing binaries in site-packages
        (wheel_dir / f'{package_name}.pth').write_text(package_name)

        # copy Python extension
        self.copy_file(extension_path, package_root)

        if deploy_dependencies:
            deploy_path_resctriction = deploy_path_resctriction or ''

            # deploy dependencies
            cmake_dependency_args = [
                'cmake',
                f'-DBUILD_EXTENSION_PATH={extension_path}',
                f'-DPACKAGE_EXTENSION_PATH={package_root / extension_path.name}',
                f'-DVORTEX_DIR={vortex_dir}',
                f'-DPACKAGE_ROOT={package_root}',
                f'-DPYTHON_EXE={sys.executable}',
                f'-DDEPLOY_PATH_RESTRICTION={deploy_path_resctriction}',
                '-P', str(ext.sourcedir / 'cmake' / 'deploy_dependencies.cmake'),
            ]
            subprocess.check_call(cmake_dependency_args)

        if generate_stubs:
            # generate stubs
            mypy_args = [
                sys.executable,
                '-c', 'print("-- Generating Stubs"); from mypy.stubgen import main; main()',
                '-p', ext.name,
                '-o', str(wheel_dir)
            ]
            subprocess.check_call(mypy_args, cwd=package_root)

# detect version
try:
    output = subprocess.check_output(['git', 'describe', '--tag', '--always', 'HEAD']).decode().strip()
    parts = output[1:].split('-')

    if len(parts) == 1:
        version = parts[0]
    else:
        version = f'{parts[0]}+{parts[1]}.{parts[2]}'

except subprocess.CalledProcessError:
    version = '0.0.0'

package_properties = dict(
    description='A library for building real-time OCT engines in C++ or Python.',
    long_description=Path(root / 'README.rst').read_text(),

    author='Mark Draelos',
    author_email='contact@vortex-oct.dev',

    license='BSD-3',
    url='https://www.vortex-oct.dev/',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: GPU :: NVIDIA CUDA',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers'
        'License :: OSI Approved :: BSD License',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: C++',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
    ],
    keywords=[
        'optical coherence tomography',
    ],

    install_requires=[
        'numpy'
    ],

    platforms=[os.name],

    zip_safe=False,
)

if __name__ == '__main__':

    setup(
        name='vortex-oct',
        version=version,

        **package_properties,

        cmdclass=dict(build_ext=CMakeBuild),
        ext_modules=[CMakeExtension('vortex', '.')],
    )
