#include <spdlog/logger.h>

#include <vortex/engine/source.hpp>
#include <vortex/engine/clock.hpp>
#include <vortex/engine/dispersion.hpp>

#if defined(VORTEX_ENABLE_ENGINE)
#  include <vortex/endpoint/cpu.hpp>
#  include <vortex/endpoint/cuda.hpp>
#  include <vortex/endpoint/storage.hpp>
#endif

#include <vortex/acquire.hpp>
#include <vortex/engine.hpp>
#include <vortex/process.hpp>
#include <vortex/scan.hpp>

#include <vortex-python/common.hpp>
#include <vortex-python/marker.hpp>
#include <vortex-python/scan.hpp>
#include <vortex-python/memory.hpp>
#include <vortex-python/exception.hpp>

template<typename T>
using sp = std::shared_ptr<T>;
template<typename T>
using spl = std::shared_ptr<vortex::sync::lockable<T>>;

template<typename T>
static void bind_utility_functions(py::module& m) {
    {
        m.def("find_rising_edges", [](const xt::pytensor<T, 1>& signal, size_t samples_per_second) {
            return vortex::engine::find_rising_edges(signal, samples_per_second);
        });
        m.def("find_rising_edges", [](const xt::pytensor<T, 1>& signal, size_t samples_per_second, size_t number_of_edges) {
            return vortex::engine::find_rising_edges(signal, samples_per_second, number_of_edges);
        });
        m.def("find_rising_edges", [](const xt::pyarray<T>& signal, size_t samples_per_second) {
            return vortex::engine::find_rising_edges(signal, samples_per_second);
        });
        m.def("find_rising_edges", [](const xt::pyarray<T>& signal, size_t samples_per_second, size_t number_of_edges) {
            return vortex::engine::find_rising_edges(signal, samples_per_second, number_of_edges);
        });
    }

    {
        m.def("compute_resampling", [](const vortex::engine::source_t<T>& source, size_t samples_per_second, size_t samples_per_ascan) {
            return vortex::engine::compute_resampling(source, samples_per_second, samples_per_ascan);
        });
        m.def("compute_resampling", [](const vortex::engine::source_t<T>& source, size_t samples_per_second, size_t samples_per_ascan, size_t clock_delay_samples) {
            return vortex::engine::compute_resampling(source, samples_per_second, samples_per_ascan, clock_delay_samples);
        });
    }

    m.def("dispersion_phasor", &vortex::engine::dispersion_phasor<T>);

#if defined(VORTEX_ENABLE_ALAZAR)
    {
        m.def("acquire_alazar_clock", [](const vortex::engine::source_t<T>& source, const vortex::alazar_acquisition_t::config_t& acquire_config, vortex::alazar::channel_t clock_channel) {
            return vortex::engine::acquire_alazar_clock<vortex::alazar_acquisition_t>(source, acquire_config, clock_channel);
        }, py::call_guard<py::gil_scoped_release>());
        m.def("acquire_alazar_clock", [](const vortex::engine::source_t<T>& source, const vortex::alazar_acquisition_t::config_t& acquire_config, vortex::alazar::channel_t clock_channel, sp<spdlog::logger> log) {
            return vortex::engine::acquire_alazar_clock<vortex::alazar_acquisition_t>(source, acquire_config, clock_channel, log);
        }, py::call_guard<py::gil_scoped_release>());
    }
#endif
}

#if defined(VORTEX_ENABLE_ENGINE)

template<typename C>
void bind_notify(py::class_<C, sp<C>>& c) {
    RW_VAR(update_callback);
    RW_VAR(segment_callback);
    RW_VAR(aggregate_segment_callback);
    RW_VAR(volume_callback);
    RW_VAR(scan_callback);
    RW_VAR(event_callback);
}

template<typename C>
static void setup_cuda_endpoint(py::class_<C, sp<C>>& c) {
    bind_notify(c);

    RO_ACC(tensor);
    RO_ACC(executor);

    // try to import cupy to allow access to CUDA stream
    py::module cupy;
    try {
        cupy = py::module::import("cupy");
    } catch (const py::error_already_set& e) {
        if (!e.matches(PyExc_ImportError)) {
            throw;
        }
        // continue on without cupy
    }

    if (cupy) {
        c.def_property_readonly("stream", [cupy](C& o) {
            return cupy.attr("cuda").attr("ExternalStream")(intptr_t(o.stream().handle()));
        });
    }
}

template<typename T>
static void bind_cuda_endpoints(py::module& m) {
    auto suffix = std::string(detail::dtype<T>::display_name);

    {
        using C = vortex::endpoint::stack_cuda_host_tensor<T>;
        auto c = py::class_<C, sp<C>>(m, ("StackHostTensorEndpoint" + suffix).c_str());

        c.def(py::init<sp<vortex::stack_format_executor_t>, std::vector<size_t>, sp<spdlog::logger>>(), "executor"_a, "shape"_a, "log"_a = nullptr);

        setup_cuda_endpoint(c);
    }

    {
        using C = vortex::endpoint::stack_cuda_device_tensor<T>;
        auto c = py::class_<C, sp<C>>(m, ("StackDeviceTensorEndpoint" + suffix).c_str());

        c.def(py::init<sp<vortex::stack_format_executor_t>, std::vector<size_t>, sp<spdlog::logger>>(), "executor"_a, "shape"_a, "log"_a = nullptr);
        c.def(py::init<sp<vortex::stack_format_executor_t>, spl<vortex::cuda::cuda_device_tensor_t<T>>, sp<spdlog::logger>>(), "executor"_a, "shape"_a, "log"_a = nullptr);

        setup_cuda_endpoint(c);
    }

    {
        using C = vortex::endpoint::radial_cuda_device_tensor<T>;
        auto c = py::class_<C, sp<C>>(m, ("RadialDeviceTensorEndpoint" + suffix).c_str());

        c.def(py::init<sp<vortex::radial_format_executor_t>, std::vector<size_t>, sp<spdlog::logger>>(), "executor"_a, "shape"_a, "log"_a = nullptr);
        c.def(py::init<sp<vortex::radial_format_executor_t>, spl<vortex::cuda::cuda_device_tensor_t<T>>, sp<spdlog::logger>>(), "executor"_a, "shape"_a, "log"_a = nullptr);

        setup_cuda_endpoint(c);
    }

    {
        using C = vortex::endpoint::position_cuda_device_tensor<T>;
        auto c = py::class_<C, sp<C>>(m, ("PositionDeviceTensorEndpoint" + suffix).c_str());

        c.def(py::init<sp<vortex::position_format_executor_t>, std::vector<size_t>, sp<spdlog::logger>>(), "executor"_a, "shape"_a, "log"_a = nullptr);
        c.def(py::init<sp<vortex::position_format_executor_t>, spl<vortex::cuda::cuda_device_tensor_t<T>>, sp<spdlog::logger>>(), "executor"_a, "shape"_a, "log"_a = nullptr);

        setup_cuda_endpoint(c);
    }
}

template<typename block_t, typename T>
static void bind_cuda_endpoints_adapters(py::module& m) {
    {
        using C = vortex::endpoint::stack_cuda_host_tensor<T>;
        m.def("_bind", [](sp<C>& a) { return vortex::engine::bind::template endpoint<block_t>(a); });
    }

    {
        using C = vortex::endpoint::stack_cuda_device_tensor<T>;
        m.def("_bind", [](sp<C>& a) { return vortex::engine::bind::template endpoint<block_t>(a); });
    }

    {
        using C = vortex::endpoint::radial_cuda_device_tensor<T>;
        m.def("_bind", [](sp<C>& a) { return vortex::engine::bind::template endpoint<block_t>(a); });
    }

    {
        using C = vortex::endpoint::position_cuda_device_tensor<T>;
        m.def("_bind", [](sp<C>& a) { return vortex::engine::bind::template endpoint<block_t>(a); });
    }
}


template<typename T>
static void bind_cpu_endpoints(py::module& m) {
    auto suffix = std::string(detail::dtype<T>::display_name);

    {
        using C = vortex::endpoint::stack_cpu_tensor<T>;
        auto c = py::class_<C, sp<C>>(m, ("StackTensorEndpoint" + suffix).c_str());

        c.def(py::init<sp<vortex::stack_format_executor_t>, std::vector<size_t>, sp<spdlog::logger>>(), "executor"_a, "shape"_a, "log"_a = nullptr);
        c.def(py::init<sp<vortex::stack_format_executor_t>, spl<vortex::cpu_tensor_t<T>>, sp<spdlog::logger>>(), "executor"_a, "tensor"_a, "log"_a = nullptr);

        bind_notify(c);

        RO_ACC(tensor);
        RO_ACC(executor);
    }
}

template<typename block_t, typename T>
static void bind_cpu_endpoints_adapters(py::module& m) {
    {
        using C = vortex::endpoint::stack_cpu_tensor<T>;
        m.def("_bind", [](sp<C>& a) { return vortex::engine::bind::template endpoint<block_t>(a); });
    }
}

#endif

template<typename T>
static void bind_source(py::module& m, const std::string& name) {
    using C = vortex::engine::source_t<T>;
    auto c = py::class_<C>(m, name.c_str());

    c.def(py::init<size_t, size_t, T, T>(),
        "triggers_per_second"_a = 100'000,
        "clock_rising_edges_per_trigger"_a = 1376,
        "duty_cycle"_a = 0.5,
        "imaging_depth_meters"_a = 0.01
    );

    RW_VAR(triggers_per_second);
    RW_VAR(clock_rising_edges_per_trigger);
    RW_VAR(duty_cycle);
    RW_VAR(imaging_depth_meters);

    RO_ACC(has_clock);
    RW_ACC_XT(clock_edges_seconds);

    c.def("__repr__", [](const C& o) {
        return fmt::format("Source(triggers_per_second={}, clock_rising_edges_per_second={}, duty_cycle={}, imaging_depth_meters={})", o.triggers_per_second, o.clock_rising_edges_per_trigger, o.duty_cycle, o.imaging_depth_meters);
    });

    SHALLOW_COPY();
}

template<typename engine_t, typename scan_t, typename C>
static void setup_online_scan_queue(py::class_<C, sp<C>>& c) {
    using callback_t = typename engine_t::scan_queue_t::scan_callback_t;
    c.def("append", [](C& o, sp<scan_t>& scan, callback_t callback, vortex::marker::scan_boundary marker) {
        o.append(scan, std::move(callback), std::move(marker));
    }, "scan"_a, "callback"_a = callback_t(), "marker"_a = vortex::marker::scan_boundary());
}

template<typename engine_t, typename scan_t, typename C>
static void setup_scan_queue(py::class_<C, sp<C>>& c) {
    using callback_t = typename engine_t::scan_queue_t::scan_callback_t;

    c.def("append", [](C& o, sp<scan_t>& scan, callback_t callback, vortex::marker::scan_boundary marker) {
        py::gil_scoped_release gil;
        o.append(scan, std::move(callback), std::move(marker));
    }, "scan"_a, "callback"_a = callback_t(), "marker"_a = vortex::marker::scan_boundary());

    c.def("interrupt", [](C& o, sp<scan_t>& scan, callback_t callback, vortex::marker::scan_boundary marker) {
        py::gil_scoped_release gil;
        o.interrupt(scan, std::move(callback), std::move(marker));
    }, "scan"_a, "callback"_a = callback_t(), "marker"_a = vortex::marker::scan_boundary());
}


#if defined(VORTEX_ENABLE_ENGINE)

template<typename engine_t, typename T>
static void bind_node_binder(py::module& m, const std::string& name) {
    m.def(name.c_str(), [m](py::sequence nodes) {
        auto bind = m.attr("_bind");
        T root;

        // inspect all arguments
        for (auto node : nodes) {
            try {
                // check if already bound
                auto n = node.cast<typename engine_t::config_t::node>();

                // binding is already done
                root.subgraph.push_back(n);
            } catch (const py::cast_error&) {
                // bind now
                auto v = bind(node).cast<typename engine_t::processor_t>();
                root.subgraph.push_back(v);
            }
        }

        return root;
    });
}

template<typename engine_t>
static void bind_engine_engine(py::module& m) {
    {
        using C = typename engine_t::scan_queue_t;
        using point_t = xtensor_to_pytensor_t<typename C::point_t>;
        auto c = py::class_<C, sp<C>>(m, "ScanQueue");

        c.def(py::init());
        c.def(py::init([](vortex::counter_t sample, const point_t& position, const point_t& velocity) {
            return std::make_shared<C>(sample, position, velocity);
        }));

#if defined(VORTEX_ENABLE_REFLEXXES)
        setup_scan_queue<engine_t, vortex::raster_scan_t>(c);
        setup_scan_queue<engine_t, vortex::repeated_raster_scan_t>(c);
        setup_scan_queue<engine_t, vortex::radial_scan_t>(c);
        setup_scan_queue<engine_t, vortex::repeated_radial_scan_t>(c);
        setup_scan_queue<engine_t, vortex::spiral_scan_t>(c);
        setup_scan_queue<engine_t, vortex::freeform_scan_t>(c);
#endif

        c.def("generate", [](C& o, std::vector<vortex::default_marker_t>& markers, const vortex::cpu_view_t<typename C::element_t>& buffer, bool zero_order_hold) {

            py::gil_scoped_release gil;
            return o.generate(markers, buffer, zero_order_hold);

        }, "markers"_a, "buffer"_a, "zero_order_hold"_a = true);

        c.def("reset", py::overload_cast<>(&C::reset));
        c.def("reset", [](C& o, vortex::counter_t sample, const point_t& position, const point_t& velocity) {
            o.reset(sample, position, velocity);
        }, "sample"_a, "position"_a, "velocity"_a);

        FXN(clear);

        RW_MUT(empty_callback);

        {
            using C2 = typename engine_t::scan_queue_t::online_scan_queue_t;
            auto c2 = py::class_<C2, sp<C2>>(c, "OnlineScanQueue");

#if defined(VORTEX_ENABLE_REFLEXXES)
            setup_online_scan_queue<engine_t, vortex::raster_scan_t>(c2);
            setup_online_scan_queue<engine_t, vortex::radial_scan_t>(c2);
            setup_online_scan_queue<engine_t, vortex::freeform_scan_t>(c2);
#endif
        }

        py::enum_<typename C::event_t>(c, "Event")
            .value("Start", C::event_t::start)
            .value("Finish", C::event_t::finish)
            .value("Interrupt", C::event_t::interrupt)
            .value("Abort", C::event_t::abort);
    }

    bind_node_binder<engine_t, typename engine_t::config_t::divide_t>(m, "divide");
    bind_node_binder<engine_t, typename engine_t::config_t::cycle_t>(m, "cycle");

    // shallow bindings since this functionality is hidden from the user
    py::class_<typename engine_t::config_t::divide_t>(m, "_Divide");
    py::class_<typename engine_t::config_t::cycle_t>(m, "_Cycle");

    {
        using C = typename engine_t::config_t;
        auto c = py::class_<C>(m, "EngineConfig");

        c.def(py::init());

        RW_VAR(records_per_block);
        RW_VAR(blocks_to_allocate);
        RW_VAR(preload_count);

        RW_VAR(blocks_to_acquire);
        RW_VAR(post_scan_records);

        RW_VAR(scanner_warp);
        RW_VAR(galvo_output_channels);
        RW_VAR(galvo_input_channels);

        c.def("add_acquisition", [m](C& o, py::object acquisition, py::sequence processors, bool preload, bool master) {
            auto bind = m.attr("_bind");

            auto k = bind(acquisition).cast<typename engine_t::acquisition_t>();
            auto& config = o.acquisitions[k];

            // update options
            config.preload = preload;
            config.master = master;

            // let Python handle unpacking each py::object
            using divide_t = typename engine_t::config_t::divide_t;
            auto other = m.attr("divide")(processors).cast<divide_t>();

            std::visit(vortex::overloaded{
                [&](divide_t& n) {
                    // merge with existing divide
                    std::move(other.subgraph.begin(), other.subgraph.end(), std::back_inserter(n.subgraph));
                },
                [&](auto& n) {
                    // append graph to any existing one by dividing
                    config.graph = engine_t::config_t::divide(config.graph, other);
                }
            }, config.graph);

        }, "acquisition"_a, "processors"_a, "preload"_a = true, "master"_a = true);

        c.def("add_processor", [m](C& o, py::object processor, py::sequence formatters) {
            auto bind = m.attr("_bind");

            auto k = bind(processor).cast<typename engine_t::processor_t>();
            auto& config = o.processors[k];

            for (auto formatter : formatters) {
                auto v = bind(formatter).cast<typename engine_t::formatter_t>();
                config.graph.insert(v);
            }
        });

        c.def("add_formatter", [m](C& o, py::object formatter, py::sequence endpoints) {
            auto bind = m.attr("_bind");

            auto k = bind(formatter).cast<typename engine_t::formatter_t>();
            auto& config = o.formatters[k];

            for (auto endpoint : endpoints) {
                auto v = bind(endpoint).cast<typename engine_t::endpoint_t>();
                config.graph.insert(v);
            }
        });

        c.def("add_io", [m](C& o, py::object io, bool preload, bool master, size_t lead_samples) {
            auto bind = m.attr("_bind");

            auto k = bind(io).cast<typename engine_t::io_t>();
            auto& config = o.ios[k];

            config.preload = preload;
            config.master = master;
            config.lead_samples = lead_samples;
        }, "io"_a, "preload"_a = true, "master"_a = false, "lead_samples"_a = 0);

        FXN(validate);

        SHALLOW_COPY();
    }

    {
        using C = typename engine_t::engine_status;
        auto c = py::class_<C>(m, "EngineStatus");

        RW_VAR(active);
        RW_VAR(dispatched_blocks);
        RW_VAR(inflight_blocks);
        RW_VAR(dispatch_completion);
        RW_VAR(block_utilization);

        SHALLOW_COPY();

        c.def("__repr__", [](C& o) {
            return fmt::format("EngineStatus(active={}, dispatched={}, inflight={}, complete={}, utilization={})", o.active, o.dispatched_blocks, o.inflight_blocks, o.dispatch_completion, o.block_utilization);
        });
    }

    {
        using C = engine_t;
        auto c = py::class_<C, std::shared_ptr<C>>(m, "Engine");

        c.def(py::init());
        c.def(py::init<sp<spdlog::logger>>());

        FXN_GIL(initialize);
        FXN_GIL(prepare);

        FXN_GIL(wait);
        c.def("wait_for", [](C& o, double dt) {
            return o.wait_for(std::chrono::duration_cast<std::chrono::microseconds>(vortex::seconds(dt)));
        }, py::call_guard<py::gil_scoped_release>());

        FXN_GIL(start);
        FXN_GIL(stop);
        RO_ACC(done);
        FXN(status);

        RW_ACC(scan_queue);
        RO_ACC(config);

        RW_MUT(event_callback);
        RW_MUT(job_callback);

        {
            using C = vortex::engine::event_t;
            py::enum_<C>(c, "Event")
                .value("Launch", C::launch)
                .value("Start", C::start)
                .value("Run", C::run)
                .value("Stop", C::stop)
                .value("Complete", C::complete)
                .value("Shutdown", C::shutdown)
                .value("Exit", C::exit)
                .value("Error", C::error)
                .value("Abort", C::abort);
        }

        auto engine = c;
        {
            using C = vortex::engine::timing_t;
            auto c = py::class_<C>(engine, "JobTiming");

            c.def(py::init());

            RW_VAR(create);
            RW_VAR(service);
            RW_VAR(scan);
            RW_VAR(acquire);
            RW_VAR(process);
            RW_VAR(format);
            RW_VAR(recycle);
        }

        {
            using C = vortex::engine::session_status_t;
            auto c = py::class_<C>(engine, "SessionStatus");

            c.def(py::init());

            RW_VAR(allocated);
            RW_VAR(inflight);
            RW_VAR(dispatched);
            RW_VAR(limit);

            RO_ACC(available);
            RO_ACC(utilization);
            RO_ACC(progress);
        }
    }

    // shallow bindings for adapters since this functionality is hidden from users
    {
        py::class_<typename engine_t::acquisition_t>(m, "_AcquisitionAdapter");
        py::class_<typename engine_t::processor_t>(m, "_ProcessorAdapter");
        py::class_<typename engine_t::formatter_t>(m, "_FormatterAdapter");
        py::class_<typename engine_t::endpoint_t>(m, "_EndpointAdapter");
        py::class_<typename engine_t::io_t>(m, "_IOAdapter");
    }

    {
        using C = typename engine_t::block_t;
        auto c = py::class_<C, std::shared_ptr<C>>(m, "Block");

        c.def(py::init());

        RW_VAR(id);
        RW_VAR(timestamp);
        RW_VAR(sample);
        RW_VAR(length);
        RW_VAR(markers);

        RO_VAR(galvo_target);
        RO_VAR(sample_target);
        RO_VAR(galvo_actual);
        RO_VAR(sample_actual);

        RO_VAR(strobes);

        py::enum_<typename C::stream_index_t>(c, "StreamIndex")
            .value("GalvoTarget", C::stream_index_t::galvo_target)
            .value("SampleTarget", C::stream_index_t::sample_target)
            .value("GalvoActual", C::stream_index_t::galvo_actual)
            .value("SampleActual", C::stream_index_t::sample_actual)
            .value("Strobes", C::stream_index_t::strobes);

        // TODO: bind the rest of the members
    }

}

#endif

void bind_engine(py::module& root) {
    auto m = root.def_submodule("engine");

    bind_utility_functions<double>(m);

#if defined(VORTEX_ENABLE_ENGINE)
    using engine_t = vortex::engine_t<uint16_t, int8_t, double, uInt32>;
    using block_t = engine_t::block_t;

    {
        using C = vortex::endpoint::null;
        auto c = py::class_<C, std::shared_ptr<C>>(m, "NullEndpoint");

        c.def(py::init());
    }

    bind_cpu_endpoints<int8_t>(m);
    bind_cpu_endpoints<uint16_t>(m);
    bind_cpu_endpoints<float>(m);
    bind_cuda_endpoints<int8_t>(m);
    bind_cuda_endpoints<uint16_t>(m);
    bind_cuda_endpoints<float>(m);

    py::enum_<vortex::endpoint::buffer_strategy_t>(m, "BufferStrategy")
        .value("Block", vortex::endpoint::buffer_strategy_t::none)
        .value("Segment", vortex::endpoint::buffer_strategy_t::segment)
        .value("Volume", vortex::endpoint::buffer_strategy_t::volume);

    {
        using C = vortex::endpoint::stream_dump_storage;
        auto c = py::class_<C, sp<C>>(m, "StreamDumpStorage");

        c.def(py::init<sp<vortex::stream_dump_t>, size_t>(), "storage"_a, "lead_samples"_a = 0);

        RO_ACC(storage);
    }
    {
        using C = vortex::endpoint::marker_log_storage;
        auto c = py::class_<C, sp<C>>(m, "MarkerLogStorage");

        c.def(py::init<sp<vortex::marker_log_t>>(), "storage"_a);

        RO_ACC(storage);
    }

    {
        using C = vortex::endpoint::spectra_stream_storage<typename engine_t::acquire_element_t>;
        auto c = py::class_<C, sp<C>>(m, "SpectraStreamEndpoint");

        c.def(py::init<sp<vortex::simple_stream_t<typename engine_t::acquire_element_t>>>(), "storage"_a);

        RO_ACC(storage);
    }
    {
        using C = vortex::endpoint::ascan_stream_storage<engine_t::process_element_t>;
        auto c = py::class_<C, sp<C>>(m, "AscanStreamEndpoint");

        c.def(py::init<sp<vortex::simple_stream_t<typename engine_t::process_element_t>>>(), "storage"_a);

        RO_ACC(storage);
    }

    {
        using C = vortex::endpoint::spectra_stack_storage<typename engine_t::acquire_element_t>;
        auto c = py::class_<C, sp<C>>(m, "SpectraStackEndpoint");

        c.def(py::init<sp<vortex::simple_stack_t<typename engine_t::acquire_element_t>>, sp<spdlog::logger>>(), "storage"_a, "log"_a = nullptr);
        c.def(py::init<sp<vortex::stack_format_executor_t>, sp<vortex::simple_stack_t<typename engine_t::acquire_element_t>>, vortex::endpoint::buffer_strategy_t, sp<spdlog::logger>>(), "executor"_a, "storage"_a, "buffer_strategy"_a = vortex::endpoint::buffer_strategy_t::volume, "log"_a = nullptr);

        bind_notify(c);

        RO_ACC(storage);
        RO_ACC(executor);
    }
    {
        using C = vortex::endpoint::ascan_stack_storage<typename engine_t::process_element_t>;
        auto c = py::class_<C, sp<C>>(m, "AscanStackEndpoint");

        c.def(py::init<sp<vortex::simple_stack_t<typename engine_t::process_element_t>>, sp<spdlog::logger>>(), "storage"_a, "log"_a = nullptr);
        c.def(py::init<sp<vortex::stack_format_executor_t>, sp<vortex::simple_stack_t<typename engine_t::process_element_t>>, vortex::endpoint::buffer_strategy_t, sp<spdlog::logger>>(), "executor"_a, "storage"_a, "buffer_strategy"_a = vortex::endpoint::buffer_strategy_t::volume, "log"_a = nullptr);

        bind_notify(c);

        RO_ACC(storage);
        RO_ACC(executor);
    }

    {
        using C = vortex::endpoint::spectra_hdf5_stack_storage<typename engine_t::acquire_element_t>;
        auto c = py::class_<C, sp<C>>(m, "SpectraHDF5StackEndpoint");

        c.def(py::init<sp<vortex::hdf5_stack_t<typename engine_t::acquire_element_t>>, sp<spdlog::logger>>(), "storage"_a, "log"_a = nullptr);
        c.def(py::init<sp<vortex::stack_format_executor_t>, sp<vortex::hdf5_stack_t<typename engine_t::acquire_element_t>>, vortex::endpoint::buffer_strategy_t, sp<spdlog::logger>>(), "executor"_a, "storage"_a, "buffer_strategy"_a = vortex::endpoint::buffer_strategy_t::volume, "log"_a = nullptr);

        bind_notify(c);

        RO_ACC(storage);
        RO_ACC(executor);
    }
    {
        using C = vortex::endpoint::ascan_hdf5_stack_storage<typename engine_t::process_element_t>;
        auto c = py::class_<C, sp<C>>(m, "AscanHDF5StackEndpoint");

        c.def(py::init<sp<vortex::hdf5_stack_t<typename engine_t::process_element_t>>, sp<spdlog::logger>>(), "storage"_a, "log"_a = nullptr);
        c.def(py::init<sp<vortex::stack_format_executor_t>, sp<vortex::hdf5_stack_t<typename engine_t::process_element_t>>, vortex::endpoint::buffer_strategy_t, sp<spdlog::logger>>(), "executor"_a, "storage"_a, "buffer_strategy"_a = vortex::endpoint::buffer_strategy_t::volume, "log"_a = nullptr);

        bind_notify(c);

        RO_ACC(storage);
        RO_ACC(executor);
    }

    {
        using C = vortex::endpoint::broct_storage;
        auto c = py::class_<C, sp<C>>(m, "BroctStorageEndpoint");

        c.def(py::init<sp<vortex::broct_format_executor_t>, sp<vortex::broct_storage_t>, vortex::endpoint::buffer_strategy_t, sp<spdlog::logger>>(), "executor"_a, "storage"_a, "buffer_strategy"_a = vortex::endpoint::buffer_strategy_t::volume, "log"_a = nullptr);

        bind_notify(c);

        RO_ACC(storage);
        RO_ACC(executor);
    }

    {
        bind_engine_engine<engine_t>(m);

        m.def("_bind", [](sp<vortex::null_acquisition_t<uint16_t>>& a) {
            return vortex::engine::bind::acquisition<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::file_acquisition_t<uint16_t>>& a) {
            return vortex::engine::bind::acquisition<block_t>(a);
        });
#if defined(VORTEX_ENABLE_ALAZAR)
        m.def("_bind", [](sp<vortex::alazar_acquisition_t>& a) {
            return vortex::engine::bind::acquisition<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::alazar_fft_acquisition_t>& a) {
            return vortex::engine::bind::acquisition<block_t>(a);
        });
#endif
#if defined(VORTEX_ENABLE_ALAZAR_GPU)
        m.def("_bind", [](sp<vortex::alazar_gpu_acquisition_t>& a) {
            return vortex::engine::bind::acquisition<block_t>(a);
        });
#endif
        m.def("_bind", [](sp<vortex::null_processor_t>& a) {
            return vortex::engine::bind::processor<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::copy_processor_t<engine_t::acquire_element_t, engine_t::process_element_t>>& a) {
            return vortex::engine::bind::processor<block_t>(a);
            });
        m.def("_bind", [](sp<vortex::cuda_processor_t<engine_t::acquire_element_t, engine_t::process_element_t>>& a) {
            return vortex::engine::bind::processor<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::null_io_t>& a) {
            return vortex::engine::bind::io<block_t>(a);
        });
#if defined(VORTEX_ENABLE_DAQMX)
        m.def("_bind", [](sp<vortex::daqmx_io_t>& a) {
            return vortex::engine::bind::io<block_t>(a);
        });
#endif
        m.def("_bind", [](sp<vortex::format_planner_t>& a) {
            return vortex::engine::bind::formatter<block_t>(a);
        });

        m.def("_bind", [](sp<vortex::endpoint::null>& a) {
            return vortex::engine::bind::endpoint<block_t>(a);
        });
        bind_cpu_endpoints_adapters<block_t, engine_t::process_element_t>(m);
        bind_cuda_endpoints_adapters<block_t, engine_t::process_element_t>(m);
        m.def("_bind", [](sp<vortex::endpoint::broct_storage>& a) {
            return vortex::engine::bind::endpoint<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::endpoint::spectra_stream_storage<engine_t::acquire_element_t>>& a) {
            return vortex::engine::bind::endpoint<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::endpoint::ascan_stream_storage<engine_t::process_element_t>>& a) {
            return vortex::engine::bind::endpoint<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::endpoint::spectra_stack_storage<engine_t::acquire_element_t>>& a) {
            return vortex::engine::bind::endpoint<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::endpoint::ascan_stack_storage<engine_t::process_element_t>>& a) {
            return vortex::engine::bind::endpoint<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::endpoint::spectra_hdf5_stack_storage<engine_t::acquire_element_t>>& a) {
            return vortex::engine::bind::endpoint<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::endpoint::ascan_hdf5_stack_storage<engine_t::process_element_t>>& a) {
            return vortex::engine::bind::endpoint<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::endpoint::stream_dump_storage>& a) {
            return vortex::engine::bind::endpoint<block_t>(a);
        });
        m.def("_bind", [](sp<vortex::endpoint::marker_log_storage>& a) {
            return vortex::engine::bind::endpoint<block_t>(a);
        });
    }

#endif

    bind_source<double>(m, "Source");

    auto sources = m.def_submodule("source");
    sources.attr("Axsun100k") = vortex::engine::source_t<double>{ 100'000, 1376, 0.50, 3.7e-3 }; // Axsun 100k
    sources.attr("Axsun200k") = vortex::engine::source_t<double>{ 200'000, 1024, 0.61, 3.7e-3 }; // Axsun 200k
    sources.attr("ThorlabsVCSEL400k") = vortex::engine::source_t<double>{ 400'000, 1024, 0.54, 0 }; // Thorlabs VCSEL 400k
}
