#include <vortex-python/common.hpp>

#include <vortex/io.hpp>

static void bind_null(py::module& m) {
    using T = vortex::io::null_io_t<vortex::io::null_config_t>;

    {
        using C = T::config_t;
        auto c = py::class_<C>(m, "NullConfig");

        c.def(py::init());

        FXN(validate);

        SHALLOW_COPY();
    }

    {
        using C = T;
        auto c = py::class_<C, std::shared_ptr<C>>(m, "NullIO");

        c.def(py::init());

        RO_ACC(config);

        FXN_GIL(initialize);

        FXN_GIL(prepare);
        FXN_GIL(start);
        FXN_GIL(stop);
    }
}

#if defined(VORTEX_ENABLE_DAQMX)

PYBIND11_MAKE_OPAQUE(std::vector<vortex::daqmx_io_t::config_t::channel_t>);

namespace pybind11 {
    namespace detail {
        template <typename... Ts>
        struct type_caster<vortex::io::channel_t<Ts...>> :
            variant_caster<vortex::io::channel_t<Ts...>> {};
    }
}

#endif

template<typename C>
static void setup_channel(py::class_<C>& c) {
    c.def(py::init());

    RW_VAR(stream);
    RW_VAR(channel);

    SHALLOW_COPY();
}


template<typename C>
static void setup_digital_channel(py::class_<C>& c) {
    c.def(py::init([](std::string line_name, size_t stream, size_t channel) {
        C o;
        o.stream = stream;
        o.channel = channel;
        o.line_name = std::move(line_name);
        return o;
    }), "line_name"_a, "stream"_a = 0, "channel"_a = 0);

    setup_channel(c);

    RW_VAR(line_name);
    SRO_VAR(max_bits);
}

template<typename C>
static void setup_analog_channel(py::class_<C>& c) {
    c.def(py::init([](std::string port_name, double logical_units_per_physical_unit, size_t stream, size_t channel) {
        C o;
        o.stream = stream;
        o.channel = channel;
        o.port_name = std::move(port_name);
        o.logical_units_per_physical_unit = logical_units_per_physical_unit;
        return o;
    }), "port_name"_a, "logical_units_per_physical_unit"_a = 1.0, "stream"_a = 0, "channel"_a = 0);

    setup_channel(c);

    RW_VAR(port_name);

    RW_VAR(logical_units_per_physical_unit);
    RW_VAR(limits);
}


#if defined(VORTEX_ENABLE_DAQMX)

static void bind_daqmx(py::module& m) {
    py::bind_vector<std::vector<vortex::daqmx_io_t::config_t::channel_t>>(m, "VectorChannel")
    .def("__repr__", [](const std::vector<vortex::daqmx_io_t::config_t::channel_t>& v){
                 return vector_repr(v, "VectorChannel");
             });

    {
        using C = vortex::io::channel::digital_output_t;
        auto c = py::class_<C>(m, "DigitalOutput");

        setup_digital_channel(c);
    }
    {
        using C = vortex::io::channel::digital_input_t;
        auto c = py::class_<C>(m, "DigitalInput");

        setup_digital_channel(c);
    }
    {
        using C = vortex::io::channel::analog_voltage_output_t;
        auto c = py::class_<C>(m, "AnalogVoltageOutput");

        setup_analog_channel(c);
    }
    {
        using C = vortex::io::channel::analog_voltage_input_t;
        auto c = py::class_<C>(m, "AnalogVoltageInput");

        setup_analog_channel(c);

        RW_VAR(terminal);
    }

    using T = vortex::daqmx_io_t;
    {
        using C = decltype(T::config_t::clock);
        auto c = py::class_<C>(m, "DAQmxConfigClock");

        c.def(py::init());

        RW_VAR(source);
        RW_VAR(edge);
        RW_VAR(divisor);

        SHALLOW_COPY();
    }

    {
        using C = T::config_t;
        auto c = py::class_<C>(m, "DAQmxConfig");

        c.def(py::init());

        RW_VAR(name);

        RW_ACC(samples_per_second);
        RW_ACC(samples_per_block);

        RW_VAR(clock);

        RW_VAR(blocks_to_buffer);

        RW_VAR(readwrite_timeout);

        RW_VAR(channels);

        RW_VAR(stop_on_error);

        FXN(validate);

        SHALLOW_COPY();
    }

    {
        using C = T;
        auto c = py::class_<C, std::shared_ptr<C>>(m, "DAQmxIO");

        c.def(py::init());
        c.def(py::init<std::shared_ptr<spdlog::logger>>());

        RO_ACC(config);

        FXN_GIL(initialize);

        FXN_GIL(prepare);
        FXN_GIL(start);
        FXN_GIL(stop);

        RO_ACC(running);
    }
}

#endif

void bind_io(py::module& root) {
    auto m = root.def_submodule("io");

    bind_null(m);

#if defined(VORTEX_ENABLE_DAQMX)
    bind_daqmx(m);
#endif

}
