#include <vortex-python/memory.hpp>

#include <vortex/util/sync.hpp>

using namespace std::string_literals;

template<typename C>
static void setup_base_tensor(const std::string& name, py::class_<C, std::shared_ptr<C>>& c) {

    c.def("__repr__", [name](const C& o) {
        return fmt::format("{}(shape=[{{}}], dtype={})", name, vortex::join(o.shape(), ", "), dtype<typename C::element_t>::name);
    });

    RO_ACC(data);
    c.def_property_readonly("dtype", [](const C&) { return py::dtype::of<typename C::element_t>(); });
    RO_ACC(valid);

    RO_ACC(count);
    RO_ACC(size_in_bytes);
    RO_ACC(stride_in_bytes);
    RO_ACC(underlying_size_in_bytes);

    RO_ACC(dimension);
    RO_ACC(shape);
    RO_ACC(stride);

    c.def("resize", [](C& o, std::vector<size_t> shape) {o.resize(shape); }, "shape"_a);
    FXN(shrink);
    FXN(clear);

    c.def("__exit__", [](C& o, py::object exc_type, py::object exc_value, py::object traceback) {
        o.mutex().unlock_shared();
        return false;
    });

}

template<typename T>
static void bind_cpu_tensor(py::module& m) {
    using C = vortex::sync::lockable<vortex::cpu_tensor_t<T>>;
    auto c = py::class_<C, std::shared_ptr<C>>(m, ("CpuTensor"s + dtype<T>::display_name).c_str());

    setup_base_tensor("CpuTensor", c);

    c.def(py::init());

    c.def("__enter__", [](C& o) {
        {
            py::gil_scoped_release gil;
            o.mutex().lock_shared();
        }

        return view(o);
    });

}

#if defined(VORTEX_ENABLE_CUDA)

template<typename T>
static void bind_cuda_host_tensor(py::module& m, py::object& cupy) {
    using C = vortex::sync::lockable<vortex::cuda::cuda_host_tensor_t<T>>;
    auto c = py::class_<C, std::shared_ptr<C>>(m, ("CudaHostTensor"s + dtype<T>::display_name).c_str());

    setup_base_tensor("CudaHostTensor", c);

    c.def(py::init());

    if (cupy) {
        c.def("__enter__", [cupy](C& o) {
            {
                py::gil_scoped_release gil;
                o.mutex().lock_shared();
            }

            return view(o);
        });
    }

}

template<typename T>
static void bind_cuda_device_tensor(py::module& m, py::object& cupy) {
    using C = vortex::sync::lockable<vortex::cuda::cuda_device_tensor_t<T>>;
    auto c = py::class_<C, std::shared_ptr<C>>(m, ("CudaDeviceTensor"s + dtype<T>::display_name).c_str());

    setup_base_tensor("CudaDeviceTensor", c);

    c.def(py::init());

    RO_ACC(device);

    if (cupy) {
        c.def("__enter__", [cupy](C& o) {
            {
                py::gil_scoped_release gil;
                o.mutex().lock_shared();
            }

            return view(o);
        });
    }

}

#endif

#if defined(VORTEX_ENABLE_ALAZAR_GPU)

template<typename T>
static void bind_alazar_device_tensor(py::module& m, py::object& cupy) {
    using C = vortex::sync::lockable<vortex::alazar::alazar_device_tensor_t<T>>;
    auto c = py::class_<C, std::shared_ptr<C>>(m, ("AlazarDeviceTensor"s + dtype<T>::display_name).c_str());

    setup_base_tensor("AlazarDeviceTensor", c);

    c.def(py::init());

    RO_ACC(device);
    RO_ACC(handle);

    if (cupy) {
        c.def("__enter__", [cupy](C& o) {
            {
                py::gil_scoped_release gil;
                o.mutex().lock_shared();
            }

            return view(o);
        });
    }

}

#endif

void bind_memory(py::module& root) {
    auto m = root.def_submodule("memory");

    bind_cpu_tensor<int8_t>(m);
    bind_cpu_tensor<uint16_t>(m);
    bind_cpu_tensor<float>(m);
    bind_cpu_tensor<double>(m);

#if defined(VORTEX_ENABLE_CUDA) || defined(VORTEX_ENABLE_ALAZAR_GPU)
    // try to import cupy to allow access to device tensors
    py::module cupy;
    try {
        cupy = py::module::import("cupy");
    } catch (const py::error_already_set& e) {
        if (!e.matches(PyExc_ImportError)) {
            throw;
        }
        // continue on without cupy
    }

#if defined(VORTEX_ENABLE_CUDA)
    bind_cuda_host_tensor<int8_t>(m, cupy);
    bind_cuda_host_tensor<uint16_t>(m, cupy);
    bind_cuda_host_tensor<float>(m, cupy);
    bind_cuda_host_tensor<double>(m, cupy);

    bind_cuda_device_tensor<int8_t>(m, cupy);
    bind_cuda_device_tensor<uint16_t>(m, cupy);
    bind_cuda_device_tensor<float>(m, cupy);
    bind_cuda_device_tensor<double>(m, cupy);
#endif

#if defined(VORTEX_ENABLE_ALAZAR_GPU)
    bind_alazar_device_tensor<int8_t>(m, cupy);
    bind_alazar_device_tensor<uint16_t>(m, cupy);
    bind_alazar_device_tensor<float>(m, cupy);
    bind_alazar_device_tensor<double>(m, cupy);
#endif

#endif

}