#include <vortex-python/common.hpp>
#include <vortex-python/memory.hpp>
#include <vortex-python/marker.hpp>
#include <vortex-python/scan.hpp>

#include <vortex/scan.hpp>

#if defined(VORTEX_ENABLE_REFLEXXES)
//#include <xtensor/xio.hpp>

#include <vortex/driver/motion.hpp>

template<typename C>
static void setup_warp(py::class_<C>& c) {

    c.def("forward", [](const C& o, xt::pyarray<double>& in) {
        xt::pyarray<double> out;
        o.forward(in, out);
        return out;
    }, "in"_a);
    c.def("inverse", [](const C& o, xt::pyarray<double>& in) {
        xt::pyarray<double> out;
        o.inverse(in, out);
        return out;
    }, "in"_a);

    //c.def("forward", [](const C& o, xt::pyarray<double>& in, xt::pyarray<double>& out) { o.forward(in, out); }, "in"_a, "out"_a);
    //c.def("inverse", [](const C& o, xt::pyarray<double>& in, xt::pyarray<double>& out) { o.inverse(in, out); }, "in"_a, "out"_a);

}

static auto bind_warp(py::module& root){
    auto m = root.def_submodule("warp");

    {
        using C = vortex::scan::warp::none_t;
        auto c = py::class_<C>(m, "NoWarp");

        c.def(py::init());
        c.def("__repr__", [](const C& v){return "NoWarp";});

        setup_warp<C>(c);
    }
    {
        using C = vortex::scan::warp::angular_t;
        auto c = py::class_<C>(m, "Angular");

        c.def(py::init());
        c.def(py::init<double>(), "factor"_a);
        c.def("__repr__", [](const C& v){return fmt::format("Angular(factor={})", v.factor);});

        RW_VAR(factor);
        setup_warp<C>(c);
    }
    {
        using C = vortex::scan::warp::telecentric_t;
        auto c = py::class_<C>(m, "Telecentric");

        c.def(py::init());
        c.def(py::init<double>(), "galvo_lens_spacing"_a);
        c.def(py::init<double, double>(), "galvo_lens_spacing"_a, "scale"_a);
        c.def("__repr__", [](const C& v){return fmt::format("Telecentric(galvo_lens_spacing={}, scale={})",
                                                            v.galvo_lens_spacing, v.scale);});

        RW_VAR(galvo_lens_spacing);
        RW_VAR(scale);
        setup_warp<C>(c);
    }
}

template<typename T>
static auto bind_limits(py::module& m) {
    using C = vortex::motion::limits_t<T>;
    auto c = py::class_<C>(m, "Limits");

    c.def(py::init());

    RW_VAR(position);
    RW_VAR(velocity);
    RW_VAR(acceleration);

    SHALLOW_COPY();

    return c;
}

#endif

template<typename C>
static void setup_scan_config(py::class_<C>& c) {
    RW_VAR(loop);
    RW_VAR(consolidate);

    RW_VAR(samples_per_second);
    RO_ACC(sampling_interval);

    RW_VAR(channels_per_sample);

    SHALLOW_COPY();
}

template<typename C>
static void setup_segmented_scan_config(py::class_<C>& c) {
    setup_scan_config(c);

    FXN(to_segments);

    RW_VAR(limits);
    RW_VAR(bypass_limits_check);
}

template<typename C>
static void setup_xy_waypoints(py::class_<C>& c) {
    // TODO: change to an approach that allows in-place editing for xtensor_fixed
    RW_VAR_XT_FIXED(offset);
    RW_VAR(angle);

    RW_VAR(extents);
    RW_ACC(bscan_extent);
    RW_ACC(segment_extent);
    RW_ACC(volume_extent);

    RW_VAR(shape);
    RW_ACC(ascans_per_bscan);
    RW_ACC(bscans_per_volume);
    RW_ACC(samples_per_segment);
    RW_ACC(segments_per_volume);

    RW_VAR(warp);

    FXN(to_waypoints);
}

template<typename C>
static void setup_pattern(py::class_<C>& c) {
    c.def("to_pattern", [](const C& o, const xt::pytensor<double, 3>& waypoints) {
        return o.template to_pattern<double>(waypoints);
    });
    c.def("to_pattern", [](const C& o, const std::vector<xt::pytensor<double, 2>>& waypoints) {
        return o.template to_pattern<double>(waypoints.begin(), waypoints.end());
    });

    SHALLOW_COPY();
}

template<typename C>
static void setup_sequential_pattern(py::class_<C>& c) {
    RW_VAR(flags);

    RW_VAR(bidirectional_segments);
    RW_VAR(bidirectional_volumes);

    setup_pattern(c);
}

template<typename C>
static void setup_repeated_pattern(py::class_<C>& c) {
    RW_VAR(repeat_count);
    RW_VAR(repeat_period);

    RW_VAR(repeat_strategy);

    RW_VAR(bidirectional_segments);

    setup_pattern(c);
}

template<typename C>
static void setup_scan(py::class_<C, std::shared_ptr<C>>& c) {

    c.def("next", [](C& o, const vortex::cpu_view_t<typename C::element_t>& buffer) {

        std::vector<vortex::default_marker_t> markers;
        auto n = o.next(markers, buffer);

        return std::make_tuple(n, std::move(markers));

    }, "buffer"_a);
    c.def("next", [](C& o, std::vector<vortex::default_marker_t>& markers, const vortex::cpu_view_t<typename C::element_t>& buffer) {

        return o.next(markers, buffer);

    }, "markers"_a, "buffer"_a);

    RO_ACC(config);

    c.def("prepare", [](C& o) { o.prepare(); });
    c.def("prepare", [](C& o, size_t count) { o.prepare(count); }, "count"_a);

    FXN(scan_markers);
    FXN(scan_buffer);
}

template<typename C>
static void bind_xy_scan(py::module& m, const std::string& name) {
    using config_t = typename C::config_t;
    auto c = py::class_<C, std::shared_ptr<C>>(m, name.c_str());

    setup_scan(c);

    c.def(py::init());

    FXN_GIL(initialize);

    c.def("restart", [](C& o) {
        py::gil_scoped_release gil;
        o.restart();
    });
    c.def("restart", [](C& o, vortex::counter_t sample, xt::pytensor<double, 1>& position, xt::pytensor<double, 1>& velocity, bool include_start) {
        py::gil_scoped_release gil;
        return o.restart(sample, position, velocity, include_start);
    }, "sample"_a, "position"_a, "velocity"_a, "include_start"_a);

    c.def("change", [](C& o, const config_t& config, bool restart) {
        py::gil_scoped_release gil;
        o.change(config, restart);
    }, "config"_a, "restart"_a = false);
    c.def("change", [](C& o, const config_t& config, bool restart, vortex::marker::event::eid_t event_id) {
        py::gil_scoped_release gil;
        o.change(config, restart, event_id);
    });

    FXN(scan_segments);
}

void bind_scan(py::module& root) {
    auto m = root.def_submodule("scan");

#if defined(VORTEX_ENABLE_REFLEXXES)
    bind_limits<double>(m);

    auto limits = m.def_submodule("limits");
    limits.attr("ThorLabs_GVS_5mm") = vortex::motion::limits_t<double>{ {-12.5, 12.5}, 8e3, 5e6 };  // ThorLabs, assuming square wave at 100 Hz
    limits.attr("ScannerMax_Saturn_5B") = vortex::motion::limits_t<double>{ {-27.5, 27.5}, 1e5, 1e9 }; // ScannerMax, assuming sine wave at 1500 Hz

    {
        using C = vortex::scan::segment_t<double, vortex::default_marker_t>;
        auto c = py::class_<C>(m, "Segment");

        c.def(py::init());
        c.def("__repr__", [](const C& v) {
            return fmt::format("Segment:\n{},\nentry_delta={}, exit_delta={}, markers={}\n",
                PY_REPR(v.position),
                PY_REPR(v.entry_delta),
                PY_REPR(v.exit_delta),
                PY_REPR(v.markers));
        });

        RW_VAR_XT(position);
        RO_ACC(entry_position);
        RO_ACC(exit_position);

        // TODO: fix these accessors
        //RW_VAR(entry_delta);
        //RW_VAR(exit_delta);
        FXN(entry_velocity);
        FXN(exit_velocity);

        RW_VAR(markers);

        SHALLOW_COPY();

        py::bind_vector<std::vector<C>>(m, "SegmentList")
            .def("__repr__", [](const std::vector<C>& v) {
            return vector_repr(v, "SegmentList");
        });
    }

    bind_warp(m);

    bind_xy_scan<vortex::raster_scan_t>(m, "RasterScan");
    bind_xy_scan<vortex::repeated_raster_scan_t>(m, "RepeatedRasterScan");
    bind_xy_scan<vortex::radial_scan_t>(m, "RadialScan");
    bind_xy_scan<vortex::repeated_radial_scan_t>(m, "RepeatedRadialScan");
    bind_xy_scan<vortex::spiral_scan_t>(m, "SpiralScan");
    bind_xy_scan<vortex::freeform_scan_t>(m, "FreeformScan");

    {
        using C = vortex::scan::raster_waypoints_t<double, vortex::default_warp_t>;
        auto c = py::class_<C>(m, "RasterWaypoints");

        c.def(py::init());
        setup_xy_waypoints(c);
    }
    {
        using C = vortex::raster_scan_t::config_t;
        auto c = py::class_<C>(m, "RasterScanConfig");

        c.def(py::init());
        setup_xy_waypoints(c);
        setup_sequential_pattern(c);
        setup_segmented_scan_config(c);
    }
    {
        using C = vortex::repeated_raster_scan_t::config_t;
        auto c = py::class_<C>(m, "RepeatedRasterScanConfig");

        c.def(py::init());
        setup_xy_waypoints(c);
        setup_repeated_pattern(c);
        setup_segmented_scan_config(c);
    }
    {
        using C = vortex::scan::radial_waypoints_t<double, vortex::default_warp_t>;
        auto c = py::class_<C>(m, "RadialWaypoints");

        c.def(py::init());
        setup_xy_waypoints(c);

        FXN(set_half_evenly_spaced);
        FXN(set_evenly_spaced);
        FXN(set_aiming);
    }
    {
        using C = vortex::radial_scan_t::config_t;
        auto c = py::class_<C>(m, "RadialScanConfig");

        c.def(py::init());
        setup_xy_waypoints(c);
        setup_sequential_pattern(c);
        setup_segmented_scan_config(c);

        FXN(set_half_evenly_spaced);
        FXN(set_evenly_spaced);
        FXN(set_aiming);
    }
    {
        using C = vortex::repeated_radial_scan_t::config_t;
        auto c = py::class_<C>(m, "RepeatedRadialScanConfig");

        c.def(py::init());
        setup_xy_waypoints(c);
        setup_repeated_pattern(c);
        setup_segmented_scan_config(c);

        FXN(set_half_evenly_spaced);
        FXN(set_evenly_spaced);
        FXN(set_aiming);
    }

    {
        using C = vortex::scan::spiral_waypoints_t<double, vortex::default_warp_t>;
        auto c = py::class_<C>(m, "SpiralWaypoints");

        c.def(py::init());
        setup_xy_waypoints(c);

        RW_VAR(cav_omega);
        RW_VAR(clv_velocity);
        RW_VAR(amplitude);
        RW_VAR(pitch);

        RW_VAR(r_transition);
        RW_VAR(t_transition);
        RW_VAR(delta_t);
        RW_VAR(t_end);

        FXN(set_hybrid_spiral);
    }
    {
        using C = vortex::spiral_scan_t::config_t;
        auto c = py::class_<C>(m, "SpiralScanConfig");

        c.def(py::init());
        setup_xy_waypoints(c);
        setup_sequential_pattern(c);
        setup_segmented_scan_config(c);

        RW_VAR(cav_omega);
        RW_VAR(clv_velocity);
        RW_VAR(amplitude);
        RW_VAR(pitch);

        RW_VAR(r_transition);
        RW_VAR(t_transition);
        RW_VAR(delta_t);
        RW_VAR(t_end);

        FXN(set_hybrid_spiral);
    }

    {
        using C = vortex::freeform_scan_t::config_t;
        auto c = py::class_<C>(m, "FreeformScanConfig");

        c.def(py::init());

        setup_segmented_scan_config(c);

        RW_VAR(pattern);
    }

    {
        using C = vortex::scan::sequential_pattern_t<vortex::default_marker_t>;
        auto c = py::class_<C>(m, "SequentialPattern");

        c.def(py::init());

        setup_sequential_pattern(c);
    }

    {
        using C = vortex::scan::repeated_pattern_t<vortex::default_marker_t>;
        auto c = py::class_<C>(m, "RepeatedPattern");

        c.def(py::init());

        setup_repeated_pattern(c);

        using P = C;
        auto& p = c;
        {
            using C = typename P::repeat_pack;
            auto c = py::class_<C>(m, "RepeatPack");

            c.def(py::init());

            RW_VAR(flags);

            SHALLOW_COPY();
        }
        {
            using C = typename P::repeat_flags;
            auto c = py::class_<C>(m, "RepeatFlags");

            c.def(py::init());

            RW_VAR(flags);

            SHALLOW_COPY();
        }
    }

#endif
}
