#pragma once

#include <complex>

#include <fftw3.h>

#include <fmt/format.h>

#include <vortex/util/cast.hpp>

namespace vortex::fft {

    template<typename T>
    class fftw_plan_t {
    public:
        using real_t = T;
        using complex_t = std::complex<T>;

        fftw_plan_t() {

        }

        ~fftw_plan_t() {
            destroy();
        }

        fftw_plan_t(const fftw_plan_t&) = delete;
        fftw_plan_t& operator=(const fftw_plan_t&) = delete;

        fftw_plan_t(fftw_plan_t&& o) : fftw_plan_t() {
            *this = std::move(o);
        }
        fftw_plan_t& operator=(fftw_plan_t&& o) {
            destroy();
            _reset();

            std::swap(_plan, o._plan);

            return *this;
        }

        void forward(int count, const std::pair<int, int> stride, const std::pair<int, int>& offset, const std::vector<int>& n, complex_t* in, complex_t* out, int flags = FFTW_PRESERVE_INPUT | FFTW_MEASURE) {
            create(true, count, stride, offset, n, in, out, flags);
        }
        void inverse(int count, const std::pair<int, int> stride, const std::pair<int, int>& offset, const std::vector<int>& n, complex_t* in, complex_t* out, int flags = FFTW_PRESERVE_INPUT | FFTW_MEASURE) {
            create(false, count, stride, offset, n, in, out, flags);
        }

        void create(bool forward, int count, const std::pair<int, int> stride, const std::pair<int, int>& offset, const std::vector<int>& n, complex_t* in, complex_t* out, int flags = FFTW_PRESERVE_INPUT | FFTW_MEASURE) {
            destroy();
            _plan = fftw_plan_many_dft(downcast<int>(n.size()), n.data(), count, reinterpret_cast<fftw_complex*>(in), NULL, stride.first, offset.first, reinterpret_cast<fftw_complex*>(out), NULL, stride.second, offset.second, forward ? FFTW_FORWARD : FFTW_BACKWARD, flags);
            if (!_plan) {
                throw std::runtime_error(fmt::format("planner failed for c2c: {}", _shape_to_string(n)));
            }
        }

        void execute() {
            if (!_plan) {
                throw std::runtime_error("attempted to execute invalid plan");
            }

            fftw_execute(_plan);
        }

        void execute(complex_t* in, complex_t* out) {
            if (!_plan) {
                throw std::runtime_error("attempted to execute invalid plan");
            }

            fftw_execute_dft(_plan, reinterpret_cast<fftw_complex*>(in), reinterpret_cast<fftw_complex*>(out));
        }

        void destroy() {
            if (_plan) {
                fftw_destroy_plan(_plan);
            }

            _reset();
        }

        bool valid() const {
            return _plan != nullptr;
        }

    protected:

        std::string _shape_to_string(const std::vector<int>& n) {
            return std::accumulate(n.begin(), n.end(), std::string(), [](const std::string& a, int b) {
                if (a.empty()) {
                    return fmt::format("{}", b);
                } else {
                    return fmt::format("{} x {}", a, b);
                }
            });
        }

        void _reset() {
            _plan = nullptr;
        }

        fftw_plan _plan = nullptr;

    };

}
