/** \rst

    CUDA kernels to support position rectification

 \endrst */

#define _USE_MATH_DEFINES
#include <math.h>

#include <array>

#include <vortex/driver/cuda/types.hpp>
#include <vortex/driver/cuda/runtime.hpp>
#include <vortex/driver/cuda/kernels.cuh>

#include <vortex/format/cuda.hpp>

using namespace vortex;

template<typename in_t, typename out_t, typename factor_t, typename position_t>
__global__
static void _position_rectify(
    factor_t scale,
    factor_t offset,
    const in_t* block,
    ulonglong2 block_shape,
    longlong2 block_stride,
    const position_t* position,
    ulonglong2 position_shape,
    longlong2 position_stride,
    ulonglong2 position_channels,
    out_t* volume,
    ulonglong3 volume_shape,
    longlong3 volume_stride,
    format::detail::matrix_t<factor_t, 2, 3> transform
) {

    // moving through block
    auto bidx = blockIdx.x * blockDim.x + threadIdx.x;
        
    // check valid block coordinates
    if (bidx >= block_shape.x) {
        return;
    }

    // lookup position for this index
    const auto& xg = position[bidx * position_stride.x + position_channels.x * position_stride.y];
    const auto& yg = position[bidx * position_stride.x + position_channels.y * position_stride.y];
    
    // map position into volume coordinates
    longlong2 vidx = {
        cuda::kernel::clip_floor_cast<long long int>(transform.v[0][0] * xg + transform.v[0][1] * yg + transform.v[0][2]),
        cuda::kernel::clip_floor_cast<long long int>(transform.v[1][0] * xg + transform.v[1][1] * yg + transform.v[1][2])
    };
    
    // check valid volume coordinates
    if (vidx.x < 0 || vidx.x >= volume_shape.x || vidx.y < 0 || vidx.y >= volume_shape.y) {
        return;
    }

    // load this A-scan in the volume
    for (int z = 0; z < volume_shape.z; z++) {
        // index into volume
        auto& out = volume[vidx.x * volume_stride.x + vidx.y * volume_stride.y + z * volume_stride.z];

        // out-of-place linear transform
        out = cuda::kernel::clip_round_cast<out_t>(scale * block[bidx * block_stride.x + z * block_stride.y] + offset);
    }
}

template<typename in_t, typename out_t, typename factor_t, typename position_t>
static void _position_rectify_internal(
    const cuda::stream_t& stream,
    factor_t scale, factor_t offset,
    const in_t* block,
    size_t block_offset,
    std::array<size_t, 2> block_shape,
    std::array<ptrdiff_t, 2> block_stride,
    const position_t* position,
    size_t position_offset, 
    std::array<size_t, 2> position_shape,
    std::array<ptrdiff_t, 2> position_stride,
    std::array<size_t, 2> position_channels,
    out_t* volume,
    size_t volume_offset,
    std::array<size_t, 3> volume_shape,
    std::array<ptrdiff_t, 3> volume_stride,
    format::detail::matrix_t<factor_t, 2, 3> transform
) {
    auto threads = cuda::kernel::threads_from_shape(block_shape[0]);
    auto blocks = cuda::kernel::blocks_from_threads(threads, block_shape[0]);

    _position_rectify << <blocks, threads, 0, stream.handle() >> > (
        scale, offset,
        block + block_offset,
        { block_shape[0], block_shape[1] },
        { block_stride[0], block_stride[1] },
        position + position_offset,
        { position_shape[0], position_shape[1] },
        { position_stride[0], position_stride[1] },
        { position_channels[0], position_channels[1] },
        volume + volume_offset,
        { volume_shape[0], volume_shape[1], volume_shape[2] },
        { volume_stride[0], volume_stride[1], volume_stride[2] },
        transform
    );
#if defined(VORTEX_SERIALIZE_CUDA_KERNELS)
    cudaDeviceSynchronize();
#endif
    auto error = cudaGetLastError();
    cuda::detail::handle_error(error, "position rectify kernel launch failed: {}", cudaGetErrorName(error));
}

namespace vortex {
    namespace format {
        namespace detail {

            void position_rectify(
                const cuda::stream_t& stream, float scale, float offset,
                const int8_t* block, size_t block_offset, std::array<size_t, 2> block_shape, std::array<ptrdiff_t, 2> block_stride,
                const double* position, size_t position_offset, std::array<size_t, 2> position_shape, std::array<ptrdiff_t, 2> position_stride, std::array<size_t, 2> position_channels,
                int8_t* volume, size_t volume_offset, std::array<size_t, 3> volume_shape, std::array<ptrdiff_t, 3> volume_stride,
                matrix_t<float, 2, 3> transform
            ) {
                _position_rectify_internal(
                    stream, scale, offset,
                    block, block_offset, block_shape, block_stride,
                    position, position_offset, position_shape, position_stride, position_channels,
                    volume, volume_offset, volume_shape, volume_stride,
                    transform
                );
            }
            void position_rectify(
                const cuda::stream_t& stream, float scale, float offset,
                const float* block, size_t block_offset, std::array<size_t, 2> block_shape, std::array<ptrdiff_t, 2> block_stride,
                const double* position, size_t position_offset, std::array<size_t, 2> position_shape, std::array<ptrdiff_t, 2> position_stride, std::array<size_t, 2> position_channels,
                int8_t* volume, size_t volume_offset, std::array<size_t, 3> volume_shape, std::array<ptrdiff_t, 3> volume_stride,
                matrix_t<float, 2, 3> transform
            ) {
                _position_rectify_internal(
                    stream, scale, offset,
                    block, block_offset, block_shape, block_stride,
                    position, position_offset, position_shape, position_stride, position_channels,
                    volume, volume_offset, volume_shape, volume_stride,
                    transform
                );
            }
        }
    }
}
