#pragma once

#include <xtensor/xmath.hpp>

#include <vortex/scan/waypoints.hpp>

// Hybrid spiral scan generation
//
// This file handles generation of hybrid spiral scanning patterns,
// which are an Archimedes spiral wit a particular time
// parameterization that gives constant angular velocity (CAV)
// behavior near the middle and constant linear velocity (CLV)
// behavior towards the outside. Using a CAV approach in the middle
// region respects the bandwidth limits of the scanner, while CLV in
// the outer regions respects the slew rate limit.
//
// For more information about spiral scanning, see [1] (Izatt lab
// publication about CLV scanning) and [2] (paper from AFM describing
// hybrid scanning).
//
// [1] https://www.osapublishing.org/boe/abstract.cfm?URI=boe-9-10-5052
// [2] https://ieeexplore.ieee.org/document/7589006

namespace vortex::scan {

    template<typename T, typename warp_t>
    struct spiral_waypoints_t : detail::xy_waypoints_t<T, warp_t> {

        T cav_omega; // angular velocity
        T clv_velocity;
        T amplitude, pitch;

        // hybrid spiral things
        T r_transition, t_transition;

        // delta_t is the time shift between CAV and CLV
        // parameterizations. t_end is total spiral time. Units of
        // both are determined by the units of the parameters passed
        // to this class.
        T delta_t, t_end;

        spiral_waypoints_t() {
            set_hybrid_spiral(100, 1e3, 3, .01, 100'000);
        }

        void set_hybrid_spiral(T frequency_in, T slew_in, T amp_in, T pitch_in, size_t samples_per_second) {
            cav_omega = frequency_in * 2 * pi;
            clv_velocity = slew_in;
            amplitude = amp_in;
            pitch = pitch_in; // pardon the pun

            r_transition = clv_velocity / cav_omega;
            t_transition = 2 * pi * clv_velocity / (cav_omega * cav_omega * pitch);

            delta_t = t_transition - pi * r_transition * r_transition / (clv_velocity * pitch);

            t_end = delta_t + pi * amplitude * amplitude / (clv_velocity * pitch);

            bscans_per_volume() = 1; // one giant b-scan
            ascans_per_bscan() = t_end * samples_per_second;
        }

        auto to_waypoints() const {
            // evenly spaced interval from [0, t_end)
            auto times = xt::linspace(0.0, t_end, ascans_per_bscan());

            // prepare CAV spiral in polar coordinates
            auto r_cav = times * cav_omega * pitch / (2 * pi);
            auto theta_cav = times * cav_omega;

            // same for CLV
            auto times_shifted = times - delta_t;
            auto r_clv = xt::sqrt(clv_velocity * pitch * times_shifted / pi);
            auto theta_clv = xt::sqrt(4 * pi * clv_velocity * times_shifted / pitch);

            // determine which times are CAV or CLV. This produces an
            // array of booleans that looks like this:
            // { true ... true, false ... false }
            auto isCAV = times < t_transition;

            // merge CAV, CLV portions (xt::where behaves like the
            // ternary operator)
            auto r_hybrid = xt::where(isCAV, r_cav, r_clv);
            auto theta_hybrid = xt::where(isCAV, theta_cav, theta_clv);

            // prepare output views
            xt::xtensor<T, 3> wp({ shape[0], shape[1], 2 });
            auto xo = xt::view(wp, xt::all(), xt::all(), 0);
            auto yo = xt::view(wp, xt::all(), xt::all(), 1);

            // convert to Cartesian and apply rotation and offset
            xo = r_hybrid * xt::cos(theta_hybrid + angle) + offset(0);
            yo = r_hybrid * xt::sin(theta_hybrid + angle) + offset(1);

            // apply warp
            warp.inverse(wp);

            return wp;
        }

        using base_t = detail::xy_waypoints_t<T, warp_t>;
        using base_t::extents, base_t::shape, base_t::angle, base_t::offset, base_t::warp;
        using base_t::ascans_per_bscan, base_t::bscans_per_volume;

    };

}

#if defined(VORTEX_ENABLE_REFLEXXES)

#include <vortex/scan/pattern.hpp>

namespace vortex::scan {

    template<typename T, typename warp_t, typename marker_t>
    using spiral_config_t = detail::patterned_config_factory_t<
        T,
        spiral_waypoints_t<T, warp_t>,
        sequential_pattern_t<marker_t>
    >;

    template<typename T, typename marker_t, typename config_t>
    using spiral_scan_t = detail::patterned_scan_factory_t<T, marker_t, config_t>;

}

#endif
