import os
import sys
from tempfile import mkstemp

import pytest

import numpy
from numpy.testing import assert_allclose

import vortex

if 'hdf5' not in vortex.__feature__:
    pytest.skip('skipping due to missing HDF5 support in vortex', allow_module_level=True)

if sys.version_info < (3, 7):
    # h5py is no longer built for Python 3.6 and may have an incompatible HDF5 version
    # NOTE: import h5py after vortex for this check to be effective
    h5py = pytest.importorskip('h5py')
else:
    import h5py

@pytest.fixture(params=[(31,17,7,1), (89, 43, 12, 3)])
def shape(request):
    return request.param

@pytest.fixture
def random_data(shape):
    rng = numpy.random.default_rng(0)

    data = rng.integers(-128, 127, size=shape, dtype=numpy.int8)

    return data

@pytest.fixture
def random_path(random_data):
    (fd, path) = mkstemp()

    yield path

    os.close(fd)
    os.remove(path)

def test_hdf5_storage_matlab(random_data, random_path):
    from vortex.storage import HDF5StackInt8, HDF5StackHeader

    storage = HDF5StackInt8()
    cfg = storage.config

    cfg.path = random_path
    cfg.shape = random_data.shape
    cfg.header = HDF5StackHeader.MATLAB

    # write data
    storage.open(cfg)
    storage.write_volume(random_data)
    storage.close()

    # read data
    with h5py.File(random_path, 'r') as f:
        output_data = numpy.array(f.get('data'))

    assert_allclose(output_data, random_data[numpy.newaxis, ...])
